// VoxelRay is a demo for raytracing in a voxel space.
// It uses a (relativly fast) 3d-bresenham-Algorithm to simulate a lightray.
// Version 0
// Author: Jakob Thomsen (http://home.in.tum.de/~thomsen)
// last update: 25.11.2002

const char *Info[] = {
  "VoxelRay - fast 3d-Bresenham-Algorithm for raytracing in a voxel space.", 
  "Copyright (C) 2002 Jakob Thomsen (http://home.in.tum.de/~thomsen).",
  "",
  "This program is free software; you can redistribute it and/or",
  "modify it under the terms of the GNU General Public License",
  "as published by the Free Software Foundation; either version 2",
  "of the License, or (at your option) any later version.",
  "",
  "This program is distributed in the hope that it will be useful,",
  "but WITHOUT ANY WARRANTY; without even the implied warranty of",
  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the",
  "GNU General Public License for more details.",
  "",
  "You should have received a copy of the GNU General Public License",
  "along with this program; if not, write to the Free Software",
  "Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.",
  "",
  0
};

#include "gfx.h"
#include "rnd.h"
#include <math.h>
#include <limits.h>

#define DEBUG

// Bildformat
#define SCRW 640
#define SCRH 480
#define SCRM 32

#define RWIDTH 200
#define RHEIGHT 50
#define RDEPTH 150

#define SWIDTH 512
#define SHEIGHT 512

unsigned int room[RWIDTH][RDEPTH][RHEIGHT];
unsigned int SkyPic[SWIDTH][SHEIGHT];

#define MAXVAL ((unsigned int)~0)
typedef unsigned char byte;

/* Three-dimensional integer vector. */
typedef struct { int x,y,z; } int3d;

/* Three-dimensional float vector. */
typedef struct { float x,y,z; } flt3d;

/* Compose components to new three-dimensional integer vector. */
int3d int3d_init(int x, int y, int z) {int3d pos={x,y,z}; return pos;}

/* Compose components to new three-dimensional float vector. */
flt3d flt3d_init(float x,float y,float z) {flt3d pos={x,y,z};return pos;}

/* Multiplicate scalar with every component of float vector. */
flt3d flt3d_scale(float s,flt3d vec) {vec.x*=s;vec.y*=s;vec.z*=s;return vec;}

/* Convert float-vector to integer-vector by rounding it's components. */
int3d round3d(flt3d flpos) {return int3d_init(flpos.x, flpos.y, flpos.z);}

// fill the Voxel-Space with forms.
void initRoom()
{
  int x,y,z;
  //double d;

  for(z=0; z<RHEIGHT; z++)
    for(y=0; y<RDEPTH; y++)
      for(x=0; x<RWIDTH; x++)
      {
	room[x][y][z] = 0;
	// mark the borders of the room.
	if(y==0 || y==RDEPTH-1) room[x][y][z]=MAXVAL;
	if(z==0 || z==RHEIGHT-1) room[x][y][z]=MAXVAL;
	if(x==0 || x==RWIDTH-1) room[x][y][z]=MAXVAL;
	// Sphere
	//d = sqrt(pow(((double)x-RWIDTH/2)/((double)RWIDTH/2),2)+pow(((double)y-RDEPTH/2)/((double)RDEPTH/2),2)+pow(((double)z-RHEIGHT)/((double)RHEIGHT),2));
	//if(d<.25) room[x][y][z] = GfxEnRgb(127,127,127);
	//if(z<10.0*sin(2.0*M_PI*d)) room[x][y][z]=GfxEnRgb(255,255,255);
	// Cuboid
	//if(z<20 && x<20 && y>RDEPTH/2) room[x][y][z]=GfxEnRgb(255,255,0);
	if(x>30 && x<50 && y>30 && y<50 && z<RHEIGHT/2) room[x][y][z]=GfxEnRgb(255,255,0);
	// Pyramid
	if(x+y+z<50) room[x][y][z]=GfxEnRgb(255,255,255);

	// Cube in the corner
	if(RWIDTH-x<20 && RDEPTH-y<20 && RHEIGHT-z<20) room[x][y][z]=GfxEnRgb(0,255,255);

	// Oktaeder
	if(abs(RWIDTH-20-x)+abs(y-20)+abs(z-RHEIGHT/2)<20) room[x][y][z] = GfxEnRgb(255,0,255);
	//if(abs(x-RWIDTH/2)+abs(y-RDEPTH/2)+abs(z-RHEIGHT/2)<20) room[x][y][z] = GfxEnRgb(255,0,255);

	// Floor
	if((z==1) && (x<RWIDTH/3 || x>RWIDTH*2/3)) room[x][y][z]=GfxEnRgb(0,0,255);
	// Walls
	if((y==1 || y==RDEPTH-2)&&(x%20<10)) room[x][y][z]=GfxEnRgb(0,255,0);
	if((x==1 || x==RWIDTH-2)&&(y%20<10)) room[x][y][z]=GfxEnRgb(255,0,0);
	// Walls
	//if((y==1 || y==RDEPTH-2)&&z<RHEIGHT/2) room[x][y][z]=GfxEnRgb(0,255,0);
	//if((x==1 || x==RWIDTH-2)&&z<RHEIGHT/2) room[x][y][z]=GfxEnRgb(255,0,0);
      }
}

// Set all the voxels in a big sphere which is wrapped with a nice picture.
void initSky()
{
  int x,y;
  int val;
  
  for(y=0;y<SHEIGHT;y++)
    for(x=0;x<SWIDTH;x++)
    {
      if(y>SHEIGHT/2)
      {
	// Blue Sky
	val=255*(SHEIGHT-y)/SHEIGHT;
	SkyPic[x][y]=GfxEnRgb(val,val,255);
      }
      else
      {
	// Some pattern
	SkyPic[x][y]=GfxEnRgb(255*(SWIDTH-x)/SWIDTH, 255*x/SWIDTH,  0);
        if((x%20<10) ^ (y%20<10)) SkyPic[x][y]=GfxEnRgb(  127,  127,  127);
      }
    }
}

// Calculate the light on a voxel.
int light(int3d pos)
{
  float vx=0,vy=0,vz=0;
  // Length of light-vector must be 1 !!!
  // Light from North-East, half-high
  float lx=sqrt(1.0/3.0),ly=sqrt(1.0/3.0),lz=sqrt(1.0/3.0);
  // Light from above.
  //float lx=0,ly=0,lz=1;
  float d;
  float s;
  float f;
  unsigned char red, green, blue;
  int i,j,k;
  
  if(pos.x<1 || pos.y<1 || pos.z<1 || pos.x>RWIDTH-2 || pos.y>RDEPTH-2 || pos.z>RHEIGHT-2) return MAXVAL;

  // calculate the normal-vector of the surface in this point.
  for(i=-1;i<=1;i++)
    for(j=-1;j<=1;j++)
      for(k=-1;k<=1;k++)
	if(room[pos.x+i][pos.y+j][pos.z+k]!=0)
	{
	  vx-=i;
	  vy-=j;
	  vz-=k;
	}

  // normalize the vector
  d = sqrt(vx*vx + vy*vy + vz*vz);
  vx/=d;
  vy/=d;
  vz/=d;
  // scalarproduct of light-vector with surface-vector (= cos(angle))
  s = lx*vx + ly*vy + lz*vz;
  // The backsiede is alwasy dark.
  if(s<0) s=0;

  // Can the source of the light be seen from this position?
  // (Light from North-East, half-high)
  for(i=1;room[pos.x+i][pos.y+i][pos.z+i]==0;i++);
  // If not: this position is in the shadow.
  if(room[pos.x+i][pos.y+i][pos.z+i]!=MAXVAL) s=0;
  // (Light from above)
  //for(i=1;room[x][y][z+i]==0;i++);
  //if(room[x][y][z+i]!=MAXVAL) s=0;
  
  // Proportion of ambient to clear light.
  f = .5+.5*s;

  // Get the color of the voxel.
  GfxDeRgb(room[pos.x][pos.y][pos.z], &red, &green, &blue);
  // Return the color of the voxel, modified by the light.
  return GfxEnRgb(red*f, green*f, blue*f);
}

// Fast 3d-Bresenham-Algorithm to trace the trajectory of a photon.
// (it doesn't stop at the borders of the space,
// so these have to be marked explicidly by the value MAXVAL)
int3d lightRay(int3d pos, int3d dir)
{
  // Loopcounter
  int i;
  // absolute values of the differences
  int dxabs,dyabs,dzabs;
  // Direction of the step
  int sdx,sdy,sdz;
  // internal...
  int x,y,z;

#ifdef debug
  // Is the startposition valid?
  //printf("%d %d %d\n", xstart, ystart, zstart);
  if (xstart<0 || xstart>=RWIDTH) printf("Startpos out of range!\n");
  if (ystart<0 || ystart>=RDEPTH) printf("Startpos out of range!\n");
  if (zstart<0 || zstart>=RHEIGHT) printf("Startpos out of range!\n");
  //printf("...\n");
#endif

  // absolute values of the differences
  dxabs=abs(dir.x);
  dyabs=abs(dir.y);
  dzabs=abs(dir.z);

  // Direction of the step (1 or -1)
  if (dir.x>=0) sdx = 1; else sdx = -1;
  if (dir.y>=0) sdy = 1; else sdy = -1;
  if (dir.z>=0) sdz = 1; else sdz = -1;

  // XXX This is not really correct, but it seems to work...
  x=0;//dyabs>>1;
  y=0;//dzabs>>1;
  z=0;//dxabs>>1;

  // interpret ray as a funktion of the "longest coordinate"
  if (dxabs>=dyabs && dxabs>=dzabs)
  {
    // Way along the X-coordinate is the longest.
    // Walk along the X-coordinate.
    for(;;)
    {
      y+=dyabs;
      z+=dzabs;
      if (y>=dxabs)
      {
        y-=dxabs;
        pos.y+=sdy;
      }
      if (z>=dxabs)
      {
        z-=dxabs;
        pos.z+=sdz;
      }
      pos.x+=sdx;

      // Is the current voxel opaque?
      if(room[pos.x][pos.y][pos.z]!=0) break;
    }
  }
  else if (dyabs>=dzabs)
  {
    // Way along the Y-coordinate is the longest.
    // Walk along the Y-coordinate.
    for(;;)
    {
      x+=dxabs;
      z+=dzabs;
      if (x>=dyabs)
      {
	x-=dyabs;
	pos.x+=sdx;
      }
      if (z>=dyabs)
      {
	z-=dyabs;
	pos.z+=sdz;
      }
      pos.y+=sdy;
      
      // Is the current voxel opaque?
      if(room[pos.x][pos.y][pos.z]!=0) break;
    }
  }
  else
  {
    // Way along the Z-coordinate is the longest.
    // Walk along the Z-coordinate.
    for(i=0;i<dzabs;i++)
    {
      x+=dxabs;
      y+=dyabs;
      if (x>=dzabs)
      {
	x-=dzabs;
	pos.x+=sdx;
      }
      if (y>=dzabs)
      {
	y-=dzabs;
	pos.y+=sdy;
      }
      pos.z+=sdz;
      
      // Is the current voxel opaque?
      if(room[pos.x][pos.y][pos.z]!=0) break;
    }
  }
  // return the position of first opaque voxel reached.
  return pos;
}

// Compute and draw the view of the scene.
// Trace every lightray back to the voxel it came from.
void view(int3d start, int alpha, int beta)
{
  int a,b;
  float A,B;
  int value;
  float tX, tY, tZ;
  float Px, Py, Pz;
  float Vx, Vy, Vz;
  float Ux, Uy, Uz;
  int mod360(int phi) {return div(div(phi,360).rem + 360, 360).rem;}
  int xSky, ySky;
  float d;
  float normX, normY, normZ;
  int3d pos, dir;

#ifdef DEBUG
  alpha=mod360(alpha);
  beta = mod360(beta);
#endif

  // All three vectors are normalized (length 1).
  // Viewpoint
  Px = sin(M_PI*alpha/180.0)*cos(M_PI*beta/180.0);
  Py = cos(M_PI*alpha/180.0)*cos(M_PI*beta/180.0);
  Pz =                       sin(M_PI*beta/180.0);
  // direction u
  Ux = cos(M_PI*alpha/180.0);
  Uy = (-sin(M_PI*alpha/180.0));
  Uz = 0;
  // direction v
  Vx = sin(M_PI*alpha/180.0)*(-sin(M_PI*beta/180.0));
  Vy = cos(M_PI*alpha/180.0)*(-sin(M_PI*beta/180.0));
  Vz =                       cos(M_PI*beta/180.0);

  for (b=0; b<SCRH; b+=1)
    for (a=0; a<SCRW; a+=1)
    {
      // View is a pyramid with a slope of 1 (=45 DEG).
      // -1 <= A,B <= +1
      A = 1.0*(double)(a-SCRW/2)/(double)(SCRW/2);
      B = 1.0*(double)(b-SCRH/2)/(double)(SCRH/2);

      // Direction of current lightray.
      tX = Px + A*Ux + B*Vx;
      tY = Py + A*Uy + B*Vy;
      tZ = Pz + A*Uz + B*Vz;
    
      dir.x=tX*1000;dir.y=tY*1000;dir.z=tZ*1000;

      // Calculate target-voxel position using ray-tracing.
      pos = lightRay(start, dir);

      // Get the color of the voxel reached before.
      value = light(pos);

      // If the border was reached: draw the SkyShpere.
      if(value==MAXVAL)
      {
	// normalize vector of current lightray.
	d = sqrt(tX*tX+tY*tY+tZ*tZ);
	normX = tX / d;
	normY = tY / d;
	normZ = tZ / d;

	// convert it into polar coordinates (0<xSky<360, -90<ySky<90)
	xSky = 180.0*atan2(normY, normX)/M_PI;
	ySky = 180.0* asin(normZ)/M_PI;

	// Wrap the background-picture on a sphere.
	value = SkyPic[(SWIDTH-1)*(180+xSky)/360][(SHEIGHT-1)*(ySky+90)/180];
      }
     
      // Draw the pixel.
      //GfxFullBox(a,SCRH-b,a+4,SCRH-(b+4), value);
      GfxPSet(a,SCRH-b, value);
      //GfxUpdate(a,b,1,1);
    }
  GfxUpdateAll();
}

// Alternative view.
void viewold(int3d start, int alpha, int beta)
{
  int a,b;
  int A,B;
  int value;
  int tX, tY, tZ;
  int mod360(int phi) {return div(div(phi,360).rem + 360, 360).rem;}
  int3d pos, dir;

  //printf("%d\n", alpha);
  //alpha = div(alpha, 360).rem;
  //if (alpha<0) alpha+=360;
  alpha=mod360(alpha);
  //printf("%d\n", alpha);
  //beta = div(beta, 360).rem;
  beta = mod360(beta);
  if (beta<0) beta+=360;

  for (b=0; b<SCRH; b+=4)
    for (a=0; a<SCRW; a+=4)
    {
      /*
      tX = x + 2*(RWIDTH*(a-SCRW/2))/(SCRW/2);
      tY = y + 2*RDEPTH;
      tZ = z + 2*(RHEIGHT*(b-SCRH/2))/(SCRH/2);
      */
      // Werte von -90 bis +90
      A = 90.0*(double)(a-SCRW/2)/(double)(SCRW/2);
      B = 90.0*(double)(b-SCRH/2)/(double)(SCRH/2);

      //tX = x + 2.0*RHEIGHT*sin(M_PI*(alpha+A)/180.0);
      //tY = y + 2.0*RHEIGHT*cos(M_PI*(alpha+A)/180.0);
      //tZ = z + 2*(RHEIGHT*(b-SCRH/2))/(SCRH/2);
      tX = 2.0*RHEIGHT*sin(M_PI*(alpha+A)/180.0)*cos(M_PI*(beta+B)/180.0);
      tY = 2.0*RHEIGHT*cos(M_PI*(alpha+A)/180.0)*cos(M_PI*(beta+B)/180.0);
      tZ = 2.0*RHEIGHT*sin(M_PI*(beta+B)/180.0);
      
      dir.x=tX*1000;dir.y=tY*1000;dir.z=tZ*1000;
      pos = lightRay(start, dir);
      value = light(pos);
      if(value==MAXVAL)
      {
	//if(fmod(A+alpha,30)<0) value=GfxEnRgb(0,255,255);
	//else value=GfxEnRgb(0,0,0);
	//value=GfxEnRgb(0,0,255*mod360(beta+B)/360);
	value=GfxEnRgb(127,127,127);
      }
      
      GfxFullBox(a,SCRH-b,a+4,SCRH-(b+4), value);
      //GfxPSet(a,b, value);
      //GfxUpdate(a,b,1,1);
    }
  GfxUpdateAll();
}

// Change longitude of viewer, -180 <= phi < 180 
void changeLongitude(int *phi, int dphi)
{
  (*phi)+=dphi;
  if((*phi)<-180) (*phi)+=360;
  else if((*phi)>=180) (*phi)-=360;
}

// Change latitude of viewer, -90 < phi < 90
void changeLatitude(int *phi, int dphi)
{
  (*phi)+=dphi;
  if((*phi)<=-90) (*phi)=-90;
  else if((*phi)>=90) (*phi)=90;
}

// Walk forward/backward
void walkStraight(flt3d *flpos, int alpha, int step)
{
  (flpos->x)+=(float)step*sin(M_PI*alpha/180.0);
  (flpos->y)+=(float)step*cos(M_PI*alpha/180.0);
  if((flpos->x)<1) (flpos->x)=1;
  if((flpos->y)<1) (flpos->y)=1;
  if((flpos->x)>=RWIDTH-1) (flpos->x)=RWIDTH-2;
  if((flpos->y)>=RDEPTH-1) (flpos->y)=RDEPTH-2;
}

// Walk Left/Right
void walkSide(flt3d *flpos, int alpha, int step)
{
  (flpos->x)-=(float)step*cos(M_PI*alpha/180.0);
  (flpos->y)+=(float)step*sin(M_PI*alpha/180.0);
  if((flpos->x)<1) (flpos->x)=1;
  if((flpos->y)<1) (flpos->y)=1;
  if((flpos->x)>=RWIDTH-1) (flpos->x)=RWIDTH-2;
  if((flpos->y)>=RDEPTH-1) (flpos->y)=RDEPTH-2;
}

// Go up/down
void changeHeight(flt3d *flpos, int step)
{
  (flpos->z)+=step;
  if((flpos->z)<1) (flpos->z)=1;
  if((flpos->z)>=RHEIGHT-1) (flpos->z)=RHEIGHT-2;
}

// Start the demo
int main(int argc, char *argv[])
{
  int i;
  flt3d flpos;
  int alpha, beta;
  int redraw;
  SDL_Event event;//, dummy;

  for(i=0; Info[i]!=NULL; i++) printf("%s\n", Info[i]);

  GfxInit(SCRW, SCRH, SCRM);

  initRoom();
  initSky();

  /* Initialize position of viewer. */
  flpos = flt3d_init(RWIDTH/2, RDEPTH/2, RHEIGHT/2);

  alpha = 0;
  beta = 0;

  view(round3d(flpos), alpha, beta);

  printf("Keys:\n");
  printf("  cursor-keys to walk around.\n");
  printf("  '.' and '.' to rotate left/right.\n");
  printf("  <PGDOWN> and <PGUP> to look up/down.\n");
  printf("  <HOME> and <END> to walk up/down.\n");
  printf("  <SPACE> -> screendump.\n");
  
  for(;;)
  {
    redraw=0;
    //SDL_WaitEvent(&event);
    SDL_PollEvent(&event);
    // clean event queue
    //while(SDL_PollEvent(&dummy));

    switch(event.type)
    {
      case SDL_QUIT:
	exit(0);
      case SDL_MOUSEMOTION:
	/*
       	alpha+=event.motion.xrel;
       	beta+=event.motion.yrel;
	view(round3d(flpos), alpha, beta);
	*/
	break;
      case SDL_KEYDOWN:
	switch(event.key.keysym.sym)
	{
	  case SDLK_COMMA:
	    changeLongitude(&alpha, -10);
	    redraw=-1;
	    break;
	  case SDLK_PERIOD:
	    changeLongitude(&alpha, 10);
	    redraw=-1;
	    break;

	  case SDLK_UP:
	    walkStraight(&flpos, alpha, 10);
	    redraw=-1;
	    break;
	  case SDLK_DOWN:
	    walkStraight(&flpos, alpha, -10);
	    redraw=-1;
	    break;
	  case SDLK_LEFT:
	    walkSide(&flpos, alpha, 10);
	    redraw=-1;
	    break;
	  case SDLK_RIGHT:
	    walkSide(&flpos, alpha, -10);
	    redraw=-1;
	    break;

	  case SDLK_PAGEUP:
	    changeLatitude(&beta, -10);
	    redraw=-1;
	    break;
	  case SDLK_PAGEDOWN:
	    changeLatitude(&beta, 10);
	    redraw=-1;
	    break;
	  case SDLK_HOME:
	    changeHeight(&flpos, 1);
	    redraw=-1;
	    break;
	  case SDLK_END:
	    changeHeight(&flpos, -1);
	    redraw=-1;
	    break;
	  case SDLK_ESCAPE:
	    exit(0);
	  case SDLK_SPACE:
	    GfxDumpScreen("screendump.pnm");
	  default:
	    break;
	}
	break;
    }
    if(redraw) view(round3d(flpos), alpha, beta);
  }

  exit(0);
}


