#include "gfx.h"
#include <math.h>

SDL_Surface *GfxScreen;

/* forward declarations for internal functions */
inline void GfxDeRgb8(GfxColor c, unsigned char *r, unsigned char *g, unsigned char *b);
inline void GfxDeRgb15(GfxColor c, unsigned char *r, unsigned char *g, unsigned char *b);
inline void GfxDeRgb16(GfxColor c, unsigned char *r, unsigned char *g, unsigned char *b);
inline void GfxDeRgb24(GfxColor c, unsigned char *r, unsigned char *g, unsigned char *b);
inline void GfxDeRgb32(GfxColor c, unsigned char *r, unsigned char *g, unsigned char *b);

inline GfxColor GfxEnRgb8(unsigned char r, unsigned char g, unsigned char b);
inline GfxColor GfxEnRgb15(unsigned char r, unsigned char g, unsigned char b);
inline GfxColor GfxEnRgb16(unsigned char r, unsigned char g, unsigned char b);
inline GfxColor GfxEnRgb24(unsigned char r, unsigned char g, unsigned char b);
inline GfxColor GfxEnRgb32(unsigned char r, unsigned char g, unsigned char b);

inline void GfxPSet8(int x, int y, GfxColor c);
inline void GfxPSet15(int x, int y, GfxColor c);
inline void GfxPSet16(int x, int y, GfxColor c);
inline void GfxPSet24(int x, int y, GfxColor c);
inline void GfxPSet32(int x, int y, GfxColor c);

inline GfxColor GfxPGet8(int x, int y);
inline GfxColor GfxPGet15(int x, int y);
inline GfxColor GfxPGet16(int x, int y);
inline GfxColor GfxPGet24(int x, int y);
inline GfxColor GfxPGet32(int x, int y);

/* initialize SDL graphics */
void GfxInit(unsigned int width, unsigned int height, unsigned int bpp)
{
  switch(bpp)
  {
    case(8):
      GfxEnRgb = GfxEnRgb8;
      GfxDeRgb = GfxDeRgb8;
      GfxPSet = GfxPSet8;
      GfxPGet = GfxPGet8;
      break;
    case(15):
      GfxEnRgb = GfxEnRgb15;
      GfxDeRgb = GfxDeRgb15;
      GfxPSet = GfxPSet15;
      GfxPGet = GfxPGet15;
      break;
    case(16):
      GfxEnRgb = GfxEnRgb16;
      GfxDeRgb = GfxDeRgb16;
      GfxPSet = GfxPSet16;
      GfxPGet = GfxPGet16;
      break;
    case(24):
      GfxEnRgb = GfxEnRgb24;
      GfxDeRgb = GfxDeRgb24;
      GfxPSet = GfxPSet24;
      GfxPGet = GfxPGet24;
      break;
    case(32):
      GfxEnRgb = GfxEnRgb32;
      GfxDeRgb = GfxDeRgb32;
      GfxPSet = GfxPSet32;
      GfxPGet = GfxPGet32;
      break;
    default:
      fprintf(stderr, "GfxInit: %d byte per pixel not supported.\n", bpp);
      fprintf(stderr, "GfxInit: Please try 8, 15, 16, 24 or 32. \n");
      exit(0);
  }
  
  // initialize SDL
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 )
  {
    fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
    exit(1);
  }
  atexit(SDL_Quit);

  // open a window
  GfxScreen = SDL_SetVideoMode(width, height, bpp, SDL_SWSURFACE | SDL_DOUBLEBUF);
  if (GfxScreen == NULL )
  {
    fprintf(stderr, "Unable to set %dx%d video: %s\n", width, height, SDL_GetError());
    exit(1);
  }

  //SDL_WM_SetCaption("SDL", "");

  SDL_EnableUNICODE(1);
}

/* query width of screen */
int GfxWidth()
{
  return GfxScreen->w;
}

/* query height of screen */
int GfxHeight()
{
  return GfxScreen->h;
}

/* query bits per pixel of screen */
int GfxBpp()
{
  return GfxScreen->format->BitsPerPixel;
}

/* transform x and y coordinates to point */
GfxTuple GfxPoint(int x, int y)
{
  GfxTuple newPoint;

  newPoint.x = x;
  newPoint.y = y;

  return newPoint;
}

/* calculate color value rgb -> int (each channel is a value from 0 to 255) */
GfxColor GfxEnRgb8(unsigned char r, unsigned char g, unsigned char b)
{
  // 8 bits: 3 red, 3 green, 2 blue
  return (GfxColor) ((r>>5)<<5)+((g>>5)<<2)+(b>>6);
}

/* calculate color value rgb -> int (each channel is a value from 0 to 255) */
GfxColor GfxEnRgb15(unsigned char r, unsigned char g, unsigned char b)
{
  // 15 bits: 5 red, 5 green, 5 blue
  return (GfxColor) ((r>>3)<<10)+((g>>3)<<5)+(b>>3);
}

/* calculate color value rgb -> int (each channel is a value from 0 to 255) */
GfxColor GfxEnRgb16(unsigned char r, unsigned char g, unsigned char b)
{
  // 16 bits: 5 red, 6 green, 5 blue
  return (GfxColor) ((r>>3)<<11)+((g>>2)<<5)+(b>>3);
}

/* calculate color value rgb -> int (each channel is a value from 0 to 255) */
GfxColor GfxEnRgb24(unsigned char r, unsigned char g, unsigned char b)
{
  // 24 bits: 8 red, 8 green, 8 blue
  return (GfxColor) (r<<16)+(g<<8)+b;
}

/* calculate color value rgb -> int (each channel is a value from 0 to 255) */
GfxColor GfxEnRgb32(unsigned char r, unsigned char g, unsigned char b)
{
  // 32 bits: 8 red, 8 green, 8 blue
  return (GfxColor) (r<<16)+(g<<8)+b;
}

/* calculate color value int -> rgb (each channel is a value from 0 to 255) */
void GfxDeRgb8(int c, unsigned char *r, unsigned char *g, unsigned char *b)
{
  // 8 bits: 3 red, 3 green, 2 blue
  *r=((c>>5)&7)<<5;
  *g=((c>>2)&7)<<5;
  *b=(c & 3)<<6;
}

/* calculate color value int -> rgb (each channel is a value from 0 to 255) */
void GfxDeRgb15(int c, unsigned char *r, unsigned char *g, unsigned char *b)
{
  // 15 bits: 5 Bits red, 5 Bits green, 5 Bits blue
  *r=((c>>10)&31)<<3;
  *g=((c>>5)&31)<<3;
  *b=(c & 31)<<3;
}

/* calculate color value int -> rgb (each channel is a value from 0 to 255) */
void GfxDeRgb16(int c, unsigned char *r, unsigned char *g, unsigned char *b)
{
  // 16 bits: 5 red, 6 green, 5 blue
  *r=((c>>11)&31)<<3;
  *g=((c>>5)&63)<<2;
  *b=(c & 31)<<3;
}

/* calculate color value int -> rgb (each channel is a value from 0 to 255) */
void GfxDeRgb24(int c, unsigned char *r, unsigned char *g, unsigned char *b)
{
  // 24 bits: 8 red, 8 green, 8 blue
  *r=(c>>16)&0xff;
  *g=(c>>8)&0xff;
  *b=c & 0xff;
}

/* calculate color value int -> rgb (each channel is a value from 0 to 255) */
void GfxDeRgb32(int c, unsigned char *r, unsigned char *g, unsigned char *b)
{
  // 32 bits: 8 red, 8 green, 8 blue
  *r=(c>>16)&0xff;
  *g=(c>>8)&0xff;
  *b=c & 0xff;
}

/* verify if position is inside clipping area */
int GfxClip(int x, int y)
{
  if(x>=0 && y>=0 && x<GfxWidth() && y<GfxHeight()) return 0;

  return -1;
}

/* draw a point at position (x,y) with color c (in 8 bit per pixel mode) */
void GfxPSet8(int x, int y, GfxColor c)
{
  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    *((Uint8 *)GfxScreen->pixels + y*GfxScreen->pitch + x)=(unsigned char)c;
  }
}

/* draw a point at position (x,y) with color c (in 15 bit per pixel mode) */
void GfxPSet15(int x, int y, GfxColor c)
{
  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    *((Uint16 *)GfxScreen->pixels + y*GfxScreen->pitch/2 + x)=c;
  }
}

/* draw a point at position (x,y) with color c (in 16 bit per pixel mode) */
void GfxPSet16(int x, int y, GfxColor c)
{
  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    *((Uint16 *)GfxScreen->pixels + y*GfxScreen->pitch/2 + x)=c;
  }
}

/* draw a point at position (x,y) with color c (in 24 bit per pixel mode) */
void GfxPSet24(int x, int y, GfxColor c)
{
  Uint8 *bufp;

  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    bufp = (Uint8 *)GfxScreen->pixels + y*GfxScreen->pitch + x * 3;
    if(SDL_BYTEORDER == SDL_LIL_ENDIAN)
    {
      bufp[0] = c & 0xff;
      bufp[1] = (c>>8) & 0xff;
      bufp[2] = (c>>16) & 0xff;
    }
    else
    {
      bufp[2] = c & 0xff;
      bufp[1] = (c>>8) & 0xff;
      bufp[0] = (c>>16) & 0xff;
    }
  }
}

/* draw a point at position (x,y) with color c (in 32 bit per pixel mode) */
void GfxPSet32(int x, int y, GfxColor c)
{
  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    *((Uint32 *)GfxScreen->pixels + y*GfxScreen->pitch/4 + x)=c;
  }
}

/* get color of point at position(x,y) in 8 bit per pixel mode */
GfxColor GfxPGet8(int x, int y)
{
  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    return (GfxColor) (*((Uint8 *)GfxScreen->pixels + y*GfxScreen->pitch + x));
  }
  return 0;
}

/* get color of point at position(x,y) in 15 bit per pixel mode */
GfxColor GfxPGet15(int x, int y)
{
  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    return (GfxColor)(*((Uint16 *)GfxScreen->pixels+y*GfxScreen->pitch/2+x));
  }
  return 0;
}

/* get color of point at position(x,y) in 16 bit per pixel mode */
GfxColor GfxPGet16(int x, int y)
{
  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    return (GfxColor)(*((Uint16 *)GfxScreen->pixels+y*GfxScreen->pitch/2+x));
  }
  return 0;
}

/* get color of point at position(x,y) in 24 bit per pixel mode */
GfxColor GfxPGet24(int x, int y)
{
  Uint8 *bufp;

  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    bufp = (Uint8 *)GfxScreen->pixels + y*GfxScreen->pitch + x * 3;
    if(SDL_BYTEORDER == SDL_LIL_ENDIAN)
    {
      return (GfxColor) (bufp[0] + (bufp[1]<<8)+ (bufp[2]<<16));
    }
    else
    {
      return (GfxColor) (bufp[2] + (bufp[1]<<8)+ (bufp[0]<<16));
    }
  }
  return 0;
}

/* get color of point at position(x,y) in 16 bit per pixel mode */
GfxColor GfxPGet32(int x, int y)
{
  // check if point is in the screen to avoid segfaults
  if(!GfxClip(x,y))
  {
    return (GfxColor)(*((Uint32 *)GfxScreen->pixels+y*GfxScreen->pitch/4+x));
  }
  return 0;
}

/* clear screen */
void GfxCls()
{
  unsigned int x, y;

  for(y=0; y<GfxWidth(); y++)
  {
    for(x=0; x<GfxWidth(); x++)
    {
      GfxPSet(x,y,GfxEnRgb(0,0,0));
    }
  }
}

/* show the changed screen */
void GfxUpdate(int x, int y, int w, int h)
{
  SDL_UpdateRect(GfxScreen, x, y, w, h);
}

/* show the changed screen */
void GfxUpdateAll()
{
  //SDL_UpdateRect(GfxScreen, 0, 0, GfxWidth, GfxHeight);
  SDL_Flip(GfxScreen);
}

/* Bit Block Transfer */
void GfxBlt()
{
  // SDL_BLIT
  fprintf(stderr, "not implemented yet\n");
}


/* draw line with Bresenham Algorithm */
void GfxLine(int x1, int y1, int x2, int y2, int color)
{
  int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;
  int SgnInt(int nr) { if (nr==0) return 0; if (nr>0)  return 1; else return -1; }

  dx=x2-x1;
  dy=y2-y1;
  dxabs=abs(dx);
  dyabs=abs(dy);

  // TODO: clipping einbauen

  sdx=SgnInt(dx);
  sdy=SgnInt(dy);
  x=dyabs>>1;
  y=dxabs>>1;
  px=x1;
  py=y1;

  if (dxabs>=dyabs)
  {
    for(i=0;i<dxabs;i++)
    {
      y+=dyabs;
      if (y>=dxabs)
      {
        y-=dxabs;
        py+=sdy;
      }
      px+=sdx;
      GfxPSet(px,py,color);
    }
  }
  else
  {
    for(i=0;i<dyabs;i++)
    {
      x+=dxabs;
      if (x>=dyabs)
      {
        x-=dyabs;
        px+=sdx;
      }
      py+=sdy;
      GfxPSet(px,py,color);
    }
  }
}

/* draw rectangle */
void GfxEmptyBox(int startx,  int starty,  int endx,  int endy,  int color)
{
  int x,y;
  void SwapInt(int *a, int *b) { int temp; temp = *a; *a = *b; *b = temp; }

  if(startx>endx) SwapInt(&startx, &endx);
  if(starty>endy) SwapInt(&starty, &endy);

  for(y=starty; y<endy; y++)
  {
    GfxPSet(startx, y, color);
    GfxPSet(endx, y, color);
  }

  for(x=startx; x<endx; x++)
  {
    GfxPSet(x, starty, color);
    GfxPSet(x, endy, color);
  }
}

/* draw filled rectangle */
void GfxFullBox(int startx,  int starty,  int endx,  int endy,  int color)
{
  int x,y;
  void SwapInt(int *a, int *b) { int temp; temp = *a; *a = *b; *b = temp; }

  if(startx>endx) SwapInt(&startx, &endx);
  if(starty>endy) SwapInt(&starty, &endy);

  for(y=starty; y<endy; y++)
  {
    for(x=startx; x<endx; x++)
    {
      GfxPSet(x, y, color);
    }
  }
}

/* draw circle */
void GfxFullCircle(int x, int y, int radius,  int color)
{
  int i, j;
  unsigned int radius2 = radius*radius;

  for(i=-radius; i<=radius; i++)
    for(j=-radius; j<=radius; j++)
      if(i*i+j*j<radius2)
        GfxPSet(x+i, y+j, color);
}

/* create palette and fill it with rainbow colors */
GfxPalette GfxCreateRainbowPalette(int length)
{
  GfxPalette rainbow;
  int i;
  char r, g, b;

  rainbow.length = length;
  rainbow.color = (GfxColor *)malloc(length*sizeof(GfxColor));

  for(i=0;i<length;i++)
  {
    /* calculate rainbow color */
    g = (GfxColor)(127.5-127.5*cos(i*2.0*M_PI/length));
    b = (GfxColor)(127.5-127.5*cos(i*2.0*M_PI/length+2.0*M_PI/3.0));
    r = (GfxColor)(127.5-127.5*cos(i*2.0*M_PI/length+4.0*M_PI/3.0));

    /* write color to palette */
    rainbow.color[i]=GfxEnRgb(r, g, b);
  }

  /* return palette */
  return rainbow;
}

/* create palette filled with grey colors */
GfxPalette GfxCreateGreyPalette(int length)
{
  GfxPalette grey;
  int i;
  char gr;

  grey.length = length;
  grey.color = (GfxColor *)malloc(length*sizeof(GfxColor));

  for(i=0;i<length;i++)
  {
    /* calculate color */
    gr = (GfxColor)(255.0 * ((double)i)/((double)length));

    /* write color to palette */
    grey.color[i]=GfxEnRgb(gr, gr, gr);
  }

  /* return palette */
  return grey;
}

/* choose a random color */
GfxColor GfxRndColor()
{
  double rnd(double range)
  {
    return range * ((double)rand())/((double)RAND_MAX);
  }

  return GfxEnRgb(rnd(255), rnd(255), rnd(255));
}

/* save content of window to file */
int GfxDumpScreen(char* name)
{
  GfxColor p;
  unsigned char r, g, b;
  int x, y;

  FILE *target;

  target = fopen(name, "w");
  if(target==NULL)
  {
    perror("error");
    fprintf(stderr, "GfxDumpScreen: failed to open file \"%s\"!\n", name);
    return -1;
  }

  // write header
  fprintf(target, "P6\n"); // binary
  //fprintf(target, "P3\n"); // ascii
  fprintf(target, "%d %d\n", GfxWidth(), GfxHeight());
  fprintf(target, "%d\n", 255);

  for(y=0; y<GfxHeight(); y++)
  {
    for(x=0; x<GfxWidth(); x++)
    {
      p = GfxPGet(x, y);
      GfxDeRgb(p, &r, &g, &b);
      //GfxPSet(x, y, GfxEnRgb(255-r, 255-g, 255-b));
      //fprintf(target, "%d\t%d\t%d\n", r, g, b); // ascii
      fprintf(target, "%c%c%c", r,g,b); // binary
    }
  }
  fclose(target);
  return 0;
}

/* simplified isometric transformation (zoomed and centered) */
int GfxIsoX(int x, int y, int z, int zoom)
{
  return GfxWidth()/2 + (((y*zoom-x*zoom)*3)>>2);
}

/* simplified isometric transformation (zoomed and centered) */
int GfxIsoY(int x, int y, int z, int zoom)
{
  return GfxHeight()/2 - (z*zoom-((zoom*(x+y))>>1));
}

